import json, inspect
from urllib.parse import urlparse, parse_qs
from http.server import BaseHTTPRequestHandler

class RequestHandler(BaseHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        self.routes = kwargs.pop('routes', {})
        self.verbose = kwargs.pop('verbose', None)
        self.requests_count = {}
        super().__init__(*args, **kwargs)

    def do_GET(self):
        self.handle_request('GET')

    def do_POST(self):
        self.handle_request('POST')
    
    def do_PUT(self):
        self.handle_request('PUT')
    
    def do_DELETE(self):
        self.handle_request('DELETE')

    def handle_request(self, method):
        parsed_path = urlparse(self.path)
        path = parsed_path.path
        query_params = parse_qs(parsed_path.query)
        content_length = int(self.headers.get('Content-Length', 0))
        post_data = self.rfile.read(content_length) if content_length > 0 else None

        if path in self.routes:
            func, methods = self.routes[path]
            if method not in methods:
                self.send_response(405)
                self.end_headers()
                return

            try:
                if method == 'GET':
                    args = self._convert_types(func, query_params)
                elif method == 'DELETE':
                    args = self._convert_types(func, query_params)
                elif method == 'PUT':
                    data = json.loads(post_data) if post_data else {}
                    args = self._convert_types(func, data)
                else:
                    data = json.loads(post_data) if post_data else {}
                    args = self._convert_types(func, data)

                result = func(**args)
                if isinstance(result, str):
                    response = result
                    self.send_response(200)
                    self.send_header('Content-Type', 'text/html')
                else:
                    response = json.dumps({"result": result}) if not isinstance(result, dict) else json.dumps(result)
                    self.send_response(200)
                    self.send_header('Content-Type', 'application/json')
                self.end_headers()
                self.wfile.write(response.encode())
            except Exception as e:
                self.send_response(500)
                self.end_headers()
                self.wfile.write(str(e).encode())
        else:
            self.send_response(404)
            self.end_headers()

    def _convert_types(self, func, data):
        signature = inspect.signature(func)
        converted = {}
        for key, value in data.items():
            param = signature.parameters.get(key)
            if param is None:
                continue

            param_type = param.annotation
            if param_type == inspect.Parameter.empty:
                converted[key] = value[0] if isinstance(value, list) else value
            else:
                try:
                    converted[key] = param_type(value[0] if isinstance(value, list) else value)
                except ValueError as e:
                    raise ValueError(f"Incorrect type for parameter '{key}': {e}")
        
        return converted


def printc(text, color='blue'):
    """
    Print a colored text to the console.
    Parameters:
        text (str): The text to be printed.
        color (str, optional): The color of the text. Defaults to 'blue'.
    """
    c = Color()
    color_code = getattr(c, color.upper(), c.BLUE)
    print(f'{color_code}{text}{c.RESET}')

class Color:
    def __init__(self):
        self.RED = '\033[91m'
        self.GREEN = '\033[92m'
        self.YELLOW = '\033[93m'
        self.BLUE = '\033[94m'
        self.MAGENTA = '\033[95m'
        self.CYAN = '\033[96m'
        self.WHITE = '\033[97m'
        self.BLACK = '\033[30m'
        self.PURPLE = '\033[35m'
        self.ORANGE = '\033[33m'
        self.REBECCAPURPLE = '\033[45m'
        self.INDIGO = '\033[44m'
        self.TURQUOISE = '\033[36m'
        self.PINK = '\033[95m'
        self.HOTPINK = '\033[95m'
        self.RESET = '\033[0m'


import logging, warnings, os, sys
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from http.server import HTTPServer
from .request_handler import RequestHandler
from .colors import printc

logging.getLogger('werkzeug').setLevel(logging.ERROR)
warnings.filterwarnings('ignore')

class RestartHandler(FileSystemEventHandler):
    def __init__(self, restart_callback):
        self.restart_callback = restart_callback

    def on_any_event(self, event):
        # Clear terminal
        os.system('cls' if os.name == 'nt' else 'clear')
        self.restart_callback()

class Verbose:
    def __init__(self, verbose=False):
        self.verbose = verbose
    
    def debug(self, msg, color='blue'):
        if self.verbose:
            printc(f'[DEBUG] {msg}', color)

class idrc:
    def __init__(self, verbose=False):
        self.routes = {}
        self.is_verbose = verbose
        self.verbose = Verbose(self.is_verbose).debug

    def define(self, func, v=1, endpoint=None, methods=['POST']):
        if endpoint is None:
            endpoint = func.__name__
            self.verbose(f'No endpoint specified. Defaulting to function name: {endpoint}')
        
        path = f'/api/v{v}/{endpoint}'
        self.routes[path] = (func, methods)
        self.verbose(f'API endpoint registered for function `{func.__name__}`: {path}')
    
    def landing(self, html=None):
        if html is None:
            html = '<h1>My IDRC API</h1><br><p>Available endpoints:</p><ul>'
            for path in self.routes:
                html += f'<li><a href="{path}">{path}</a></li>'
            html += '</ul>'

        self.routes['/'] = (lambda: html, ['GET'])
    
    def ecode(self, code, message=''):
        self.verbose(f'Returning status code {code} with message: {message}')
        raise Exception(message)

    def run(self, *args, **kwargs):
        debug = kwargs.get('debug', False)

        if debug:
            self._run_with_reloader(*args, **kwargs)
        else:
            self._run_without_reloader(*args, **kwargs)

    def _run_with_reloader(self, *args, **kwargs):
        kwargs.pop('debug', None)

        def restart():
            printc('[INFO] Detected file change. Restarting...', 'yellow')
            os.execv(sys.executable, ['python'] + sys.argv)
        
        host = kwargs.get('host', 'localhost')
        port = kwargs.get('port', 5000)

        event_handler = RestartHandler(restart)
        observer = Observer()
        observer.schedule(event_handler, path='.', recursive=True)
        observer.start()
        printc(f'[SUCCESS] API is running on http://{host}:{port}', 'green')
 
        try:
            self._run_without_reloader(*args, **kwargs)
        finally:
            observer.stop()
            observer.join()
            self.verbose('[INFO] Observer stopped.', 'yellow')
    
    def _run_without_reloader(self, *args, **kwargs):
        host = kwargs.get('host', 'localhost')
        port = kwargs.get('port', 5000)
        available = self.port_available(host, port)
        if not available:
            printc(f'[ERROR] Port {port} is not available. Please use a different port.', 'red')
            sys.exit(1)

        error = [False]

        def start_app():
            try:
                self.landing()
                server = HTTPServer((host, port), lambda *args, **kwargs: RequestHandler(*args, routes=self.routes, verbose=self.verbose, **kwargs))
                server.serve_forever()
            except Exception as e:
                error[0] = True
                sys.stdout = sys.__stdout__
                sys.stderr = sys.__stderr__
                printc(f'[ERROR] {e}', 'red')
            finally:
                sys.stdout = sys.__stdout__
                sys.stderr = sys.__stderr__

        start_app()

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        if not error[0]:
            m = f'[SUCCESS] HTTP server running on http://{host}:{port}'
            printc(m, 'green')
            printc('[SUCCESS] Press Ctrl+C to stop the server', 'green')
            printc('='*(len(m)+10), 'yellow')

    def port_available(self, host, port):
        import socket
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex((host, port)) != 0
        